from unittest.mock import patch
from django.test import TestCase
from .generator import AbstractHtgUrlGenerator


class TestHtgUrl(TestCase):
    @patch("htg_url.generator.AbstractHtgUrlGenerator._get_from_redis")
    @patch("htg_url.generator.AbstractHtgUrlGenerator.generate_new_token")
    @patch("htg_url.generator.AbstractHtgUrlGenerator.update_ttl")
    def test_url_of_two_instances_with_same_parameters(self, mock_update_ttl,
                                                       mock_generate_new_token, mock_get_from_redis):
        class TestClass(AbstractHtgUrlGenerator):
            @staticmethod
            def create_unique_identifier(**properties):
                return '_'.join(properties.values())

        db_dict = {}

        def get_from_redis(identifier):
            return db_dict.get(identifier)

        def generate_new_token(identifier):
            token = '1234abcd5678efgh'
            db_dict.update({identifier: token})
            return token

        mock_get_from_redis.side_effect = get_from_redis
        mock_generate_new_token.side_effect = generate_new_token
        mock_update_ttl.side_effect = lambda *args: None

        test_properties = {'key1': 'test1', 'key2': 'test2', 'key3': 'test3'}

        instance = TestClass(**test_properties)
        other_instance = TestClass(**test_properties)

        instance_url = instance.url
        other_instance_url = other_instance.url

        self.assertEquals(instance_url, other_instance_url)



import base64
import io
import mimetypes
import magic
from io import BytesIO
from django.utils.module_loading import import_string
from django.conf import settings
from django.http import Http404, FileResponse
from .redis.client import RedisWrapper
from django.views import View
from wsgiref.util import FileWrapper


class AbstractDownloadDocumentView(View):
    http_method_names = ('get',)

    def get(self, request, *args, **kwargs):
        token = kwargs.get('token')
        if not token:
            raise Http404

        properties = RedisWrapper.get_dict(token)
        base_class = import_string(settings.HTG_URL_SETTINGS.get('HTG_WRAPPER_CLASS'))

        if not properties or not base_class:
            raise Http404

        instance = base_class(**properties)
        doc_identifier = instance.get_doc_identifier()

        # get document base64 string from Redis
        document_string = RedisWrapper.get(doc_identifier)
        if document_string:
            return self._get_file(document_string)

        # get document base64 string from API request
        fetcher_class = import_string(settings.HTG_URL_SETTINGS.get('DOC_WRAPPER_CLASS'))
        if not fetcher_class:
            raise Http404
        document_string = fetcher_class.fetch_document_from_sap(**properties)
        if document_string:
            RedisWrapper.set(doc_identifier, document_string)
            instance.update_ttl(token)
            return self._get_file(document_string)

        raise Http404

    @staticmethod
    def _get_file(document_string):
        buffer = BytesIO()
        buffer.write(base64.b64decode(document_string))
        buffer.seek(io.SEEK_SET)

        mime_type = magic.from_buffer(buffer.getvalue(), True)
        extension = mimetypes.guess_extension(mime_type)

        buffer.seek(io.SEEK_SET)

        return FileResponse(
            FileWrapper(buffer),
            as_attachment=True,
            filename=f'document{extension}',
            content_type=mime_type
        )



import redis
import os
from django.conf import settings


class RedisWrapper:
    _CONNECTION = None

    @classmethod
    def client(cls):
        redis_connection_string = settings.HTG_URL_SETTINGS.get('REDIS_CONNECTION_STRING')
        if cls._CONNECTION is None and redis_connection_string:
            cls._CONNECTION = redis.Redis.from_url(redis_connection_string)
            return cls._CONNECTION

        if cls._CONNECTION is None:
            cls._CONNECTION = redis.Redis(
                host=os.environ.get('REDIS_HOST'),
                port=int(os.environ.get('REDIS_PORT')),
                password=os.environ.get('REDIS_PASSWORD')
            )
        return cls._CONNECTION

    @classmethod
    def ping(cls):
        return cls.client().ping()

    @classmethod
    def get(cls, key):
        if cls.key_exists(key):
            value = cls.client().get(key)
            return value.decode("utf-8") if value else None

        return None

    @classmethod
    def get_dict(cls, hash_key):
        if cls.key_exists(hash_key):
            return {key.decode("utf-8"): value.decode("utf-8") for key, value in cls.client().hgetall(hash_key).items()}

        return None

    @classmethod
    def set(cls, key, value):
        return cls.client().set(key, value)

    @classmethod
    def set_dict(cls, name, key, value):
        return cls.client().hset(name, key, value)

    @classmethod
    def key_exists(cls, key):
        return cls.client().exists(key)

    @classmethod
    def dict_key_exists(cls, name, key):
        return cls.client().hexists(name, key)

    @classmethod
    def delete(cls, keys):
        if isinstance(keys, list):
            for key in keys:
                cls.client().delete(key)
            return True
        else:
            return cls.client().delete(keys)

    @classmethod
    def delete_dict_key(cls, name, key):
        return cls.client().hdel(name, key)

    @classmethod
    def set_ttl(cls, key, ttl=settings.HTG_URL_SETTINGS.get('HTG_URL_REDIS_TTL')):
        if cls.key_exists(key):
            return cls.client().expire(key, ttl)

        return None

    @classmethod
    def check_ttl(cls, key):
        if cls.key_exists(key):
            return cls.client().ttl(key)

        return None

    @classmethod
    def get_keys(cls, pattern='*'):
        return [key.decode('utf-8') for key in cls.client().keys(pattern)]



from htg_url.redis.client import RedisWrapper
from abc import ABC, abstractmethod


class AbstractHtgUrlGenerator(ABC):
    _HTG_PREFIX = 'HTG_'
    _DOC_PREFIX = 'DOC_'

    def __init__(self, **kwargs):
        super(AbstractHtgUrlGenerator, self).__init__()
        self.properties = kwargs
        self.unique_identifier = self.create_unique_identifier(**self.properties)

    @property
    def url(self):
        htg_identifier = self.get_htg_identifier()
        existing_token = self._get_from_redis(htg_identifier)

        if existing_token:
            self.update_ttl(existing_token)
            return existing_token

        return self.generate_new_token(htg_identifier)

    @staticmethod
    @abstractmethod
    def create_unique_identifier(**properties):
        """Get unique identifier based on passed properties
        :return: :string
        """

    def get_htg_identifier(self):
        return self._HTG_PREFIX + self.unique_identifier

    def get_doc_identifier(self):
        return self._DOC_PREFIX + self.unique_identifier

    def generate_new_token(self, unique_identifier):
        generated_token = self._generate_hex()
        self._write_to_redis(unique_identifier, generated_token)
        self._write_properties_to_redis(generated_token, self.properties)
        self.update_ttl(generated_token)
        return generated_token

    @staticmethod
    def _generate_hex():
        import secrets
        return secrets.token_hex(16)

    @staticmethod
    def _get_from_redis(key):
        return RedisWrapper.get(key)

    @staticmethod
    def _write_to_redis(key, value):
        return RedisWrapper.set(key, value)

    @staticmethod
    def _write_properties_to_redis(dict_name, data: dict):
        for key, value in data.items():
            RedisWrapper.set_dict(dict_name, key, value)

    def update_ttl(self, token):
        for key in self._get_redis_keys(token):
            RedisWrapper.set_ttl(key)

    def _get_redis_keys(self, token):
        return [self.get_htg_identifier(), self.get_doc_identifier(), token]



from abc import ABC, abstractmethod


class AbstractFetchDataFromSap(ABC):
    @staticmethod
    @abstractmethod
    def fetch_document_from_sap(**properties):
        """
        Implement logic for fetching document data from SAP and return base64 string
        :param properties:
        :return: base64 string
        """
